// ---------------------------------------------------------------------------
//  index.js
// ---------------------------------------------------------------------------

    var fs = require('fs');
    var path = require('path');

    var Promise = require('bluebird');
    var uuid = require('uuid');
    var EMPTY_STRING = require('empty-string');
    var iterateUp = require('@amphibian/iterate-up');

    var WebDriver = require('selenium-webdriver');
    var BlinkDiff = require('blink-diff');

    /**
     * Save driver screenshot
     * @param {object} driver
     * @param {string} fileName
     *
     * @returns {promise}
    **/
    function saveDriverScreenshot(driver, fileName) {
        return new Promise(function (resolve, reject) {
            return driver.takeScreenshot().then(function (data) {
                var _contents = data.replace(/^data:image\/png;base64,/, EMPTY_STRING);

                fs.writeFile(fileName, _contents, 'base64', function (error) {
                    if (error) {
                        reject(error);
                    } else {
                        resolve(fileName);
                        driver.quit();
                    }
                });
            });
        });
    }

    /**
     * Create a web driver
     * @param {string} server
     * @param {object} options
     *
     * @returns {promise}
    **/
    function createDriver(server, options) {
        return new Promise(function (resolve, reject) {
            var _driver = new WebDriver.Builder()
                .usingServer(server)
                .withCapabilities(options)
                .build();

            resolve(_driver);
        });
    }

    /**
     * Create a driver and snap a screenshot for an URL
     * @param {string} server
     * @param {object} options
     * @param {string} url
     *
     * @returns {promise}
    **/
    function snapScreenshot(server, options, url) {
        return new Promise(function (resolve, reject) {
            var _safeURL = url.replace(/[^a-zA-z0-9\_]/g, '');
            var _fileName = './screen-' + uuid.v4() + '.png';

            return createDriver(server, options).then(function (driver) {
                driver.get(url);

                // Make images invisible to avoid extreme diff count because
                // because an image is placed some pixels off
                driver.executeScript([
                    'var _images = window.document.getElementsByTagName(\'img\');',
                    'var _imagesLength = _images.length;',

                    'for (var _i = 0; _i < _imagesLength; _i++) {',
                        '_images[_i].style.visibility = \'hidden\';',
                    '}'
                ].join(''));

                // Hide CSS background images for the same reason as above
                driver.executeScript([
                    'function removeBackgrounds(element) {',
                        'if (element && element.style) {',
                            'element.style.backgroundImage = \'none\';',
                        '}',

                        'var _children = element.childNodes;',
                        'var _childrenLength = _children.length;',

                        'for (var _i = 0; _i < _childrenLength; _i++) {',
                            'removeBackgrounds(_children[_i]);',
                        '}',
                    '}',

                    'removeBackgrounds(window.document.body);'
                ].join(''));

                return resolve(saveDriverScreenshot(driver, _fileName));
            });
        });
    }

    /**
     * Snap multiple screenshots
     * @param {string} url
     * @param {array} tasks
     *
     * @returns {promise}
     * @see snapScreenshot
    **/
    function snapScreenshots(server, tasks) {
        var _promises = [];

        iterateUp(tasks, function (index) {
            _promises.push(
                snapScreenshot(server, index.options, index.url)
            );
        });

        return Promise.all(_promises);
    }

    /**
     * Compare driver screenshots
     * @param {object} options
     * @param {string} options.server
     * @param {string} options.url
     * @param {array} options.drivers
     *
     * @returns {promise}
    **/
    function compareDriverScreenshots(options) {
        return new Promise(function (resolve, reject) {
            return snapScreenshots(options.server, [{
                url: options.url,
                options: options.drivers[0]
            }, {
                url: options.url,
                options: options.drivers[1]
            }]).then(function (screenshotFileNames) {
                var _diffImagePath = './diff-' + uuid.v4() + '.png';
                var _diff = new BlinkDiff({
                    imageAPath: screenshotFileNames[0],
                    imageBPath: screenshotFileNames[1],

                    imageOutputPath: _diffImagePath
                });

                _diff.run(function (error, result) {
                    if (error) {
                        reject(error);
                    } else {
                        fs.unlink(screenshotFileNames[0]);
                        fs.unlink(screenshotFileNames[1]);

                        resolve({
                            differences: result.differences,
                            imagePath: path.resolve(_diffImagePath)
                        });
                    }
                });
            });
        });
    }

    module.exports = compareDriverScreenshots;
